r"""Polynomial flows."""

__all__ = [
    "BPF",
    "SOSPF",
]

from torch import Tensor
from torch.distributions import Transform

from .autoregressive import MAF
from ..lazy import UnconditionalTransform
from ..transforms import (
    AdditiveTransform,
    BoundedBernsteinTransform,
    ComposedTransform,
    SoftclipTransform,
    SOSPolynomialTransform,
)


def ShiftedSOSPTransform(a: Tensor, constant: Tensor) -> Transform:
    r"""Creates a shifted sum-of-squares polynomial (SOSP) transformation."""

    return ComposedTransform(
        SOSPolynomialTransform(a=a),
        AdditiveTransform(shift=constant),
    )


class SOSPF(MAF):
    r"""Creates a sum-of-squares polynomial flow (SOSPF).

    Warning:
        Invertibility is only guaranteed for features within the interval :math:`[-10,
        10]`. It is recommended to standardize features (zero mean, unit variance)
        before training.

    See also:
        :class:`zuko.transforms.SOSPolynomialTransform`

    References:
        | Sum-of-Squares Polynomial Flow (Jaini et al., 2019)
        | https://arxiv.org/abs/1905.02325

    Arguments:
        features: The number of features.
        context: The number of context features.
        degree: The degree :math:`L` of polynomials.
        polynomials: The number of polynomials :math:`K`.
        kwargs: Keyword arguments passed to :class:`zuko.flows.autoregressive.MAF`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        degree: int = 4,
        polynomials: int = 3,
        **kwargs,
    ):
        super().__init__(
            features=features,
            context=context,
            univariate=ShiftedSOSPTransform,
            shapes=[(polynomials, degree + 1), ()],
            **kwargs,
        )

        transforms = self.transform.transforms

        for i in reversed(range(1, len(transforms))):
            transforms.insert(i, UnconditionalTransform(SoftclipTransform, bound=11.0))


class BPF(MAF):
    r"""Creates a Bernstein polynomial flow (BPF).

    Warning:
        The Bernstein polynomial is bounded to the interval :math:`[-5, 5]`. Any feature
        outside of this domain is not transformed. It is recommended to standardize
        features (zero mean, unit variance) before training.

    See also:
        :class:`zuko.transforms.BoundedBernsteinTransform`

    References:
        | Deep transformation models: Tackling complex regression problems with neural network based transformation models (Sick et al., 2020)
        | https://arxiv.org/abs/2004.00464

        | Short-Term Density Forecasting of Low-Voltage Load using Bernstein-Polynomial Normalizing Flows (Arpogaus et al., 2022)
        | https://arxiv.org/abs/2204.13939

    Arguments:
        features: The number of features.
        context: The number of context features.
        degree: The degree :math:`M` of the Bernstein polynomial.
        kwargs: Keyword arguments passed to :class:`zuko.flows.autoregressive.MAF`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        degree: int = 16,
        **kwargs,
    ):
        super().__init__(
            features=features,
            context=context,
            univariate=BoundedBernsteinTransform,
            shapes=[(degree + 1,)],
            **kwargs,
        )



r"""Spline flows."""

__all__ = [
    "NCSF",
    "NSF",
]

import torch

from math import pi
from torch.distributions import Transform

from .autoregressive import MAF
from ..distributions import BoxUniform
from ..lazy import UnconditionalDistribution
from ..transforms import CircularShiftTransform, ComposedTransform, MonotonicRQSTransform


class NSF(MAF):
    r"""Creates a neural spline flow (NSF) with monotonic rational-quadratic spline
    transformations.

    By default, transformations are fully autoregressive. Coupling transformations
    can be obtained by setting :py:`passes=2`.

    Warning:
        Spline transformations are defined over the domain :math:`[-5, 5]`. Any feature
        outside of this domain is not transformed. It is recommended to standardize
        features (zero mean, unit variance) before training.

    See also:
        :class:`zuko.transforms.MonotonicRQSTransform`

    References:
        | Neural Spline Flows (Durkan et al., 2019)
        | https://arxiv.org/abs/1906.04032

    Arguments:
        features: The number of features.
        context: The number of context features.
        bins: The number of bins :math:`K`.
        kwargs: Keyword arguments passed to :class:`zuko.flows.autoregressive.MAF`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        bins: int = 8,
        **kwargs,
    ):
        super().__init__(
            features=features,
            context=context,
            univariate=MonotonicRQSTransform,
            shapes=[(bins,), (bins,), (bins - 1,)],
            **kwargs,
        )


def CircularRQSTransform(*phi) -> Transform:
    r"""Creates a circular rational-quadratic spline (RQS) transformation."""

    return ComposedTransform(
        CircularShiftTransform(bound=pi),
        MonotonicRQSTransform(*phi, bound=pi),
    )


class NCSF(MAF):
    r"""Creates a neural circular spline flow (NCSF).

    Circular spline transformations are obtained by composing circular domain shifts
    with regular spline transformations. Features are assumed to lie in the half-open
    interval :math:`[-\pi, \pi[`.

    See also:
        :class:`zuko.transforms.CircularShiftTransform`

    References:
        | Normalizing Flows on Tori and Spheres (Rezende et al., 2020)
        | https://arxiv.org/abs/2002.02428

    Arguments:
        features: The number of features.
        context: The number of context features.
        bins: The number of bins :math:`K`.
        kwargs: Keyword arguments passed to :class:`zuko.flows.autoregressive.MAF`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        bins: int = 8,
        **kwargs,
    ):
        super().__init__(
            features=features,
            context=context,
            univariate=CircularRQSTransform,
            shapes=[(bins,), (bins,), (bins - 1,)],
            **kwargs,
        )

        self.base = UnconditionalDistribution(
            BoxUniform,
            torch.full((features,), -pi - 1e-5),
            torch.full((features,), pi + 1e-5),
            buffer=True,
        )



r"""Tests for the zuko.nn module."""

import math
import pytest
import torch
import torch.nn as nn

from torch import randn
from typing import Sequence

from zuko.nn import *


@pytest.mark.parametrize("bias", [True, False])
@pytest.mark.parametrize("batch", [(), (256,)])
def test_Linear(bias: bool, batch: Sequence[int]):
    net = Linear(3, 5, bias=bias)

    x = randn(*batch, 3)
    y = net(x)

    assert y.shape == (*batch, 5)
    assert y.requires_grad


@pytest.mark.parametrize("activation", [None, torch.nn.ELU])
@pytest.mark.parametrize("normalize", [True, False])
@pytest.mark.parametrize("batch", [(), (256,)])
def test_MLP(activation: type, normalize: bool, batch: Sequence[int]):
    net = MLP(3, 5, activation=activation, normalize=normalize)

    x = randn(*batch, 3)
    y = net(x)

    assert y.shape == (*batch, 5)
    assert y.requires_grad


@pytest.mark.parametrize("residual", [True, False])
@pytest.mark.parametrize("batch", [(), (256,)])
def test_MaskedMLP(residual: bool, batch: Sequence[int]):
    adjacency = randn(5, 3) < 0
    net = MaskedMLP(adjacency, activation=nn.ELU, residual=residual)

    x = randn(*batch, 3)
    y = net(x)

    assert y.shape == (*batch, 5)
    assert y.requires_grad

    # Jacobian
    x = randn(*batch, 3)
    J = torch.autograd.functional.jacobian(net, x)
    J = J.movedim(len(batch), -2)

    mask = torch.eye(math.prod(batch), dtype=bool)
    mask = mask.reshape(batch + batch)

    assert (J[mask][..., ~adjacency] == 0).all()
    assert (J[~mask] == 0).all()


@pytest.mark.parametrize("batch", [(), (256,)])
def test_MonotonicMLP(batch: Sequence[int]):
    net = MonotonicMLP(3, 5)

    x = randn(*batch, 3)
    y = net(x)

    assert y.shape == (*batch, 5)
    assert y.requires_grad

    # Jacobian
    x = randn(*batch, 3)
    J = torch.autograd.functional.jacobian(net, x)
    J = J.movedim(len(batch), -2)

    mask = torch.eye(math.prod(batch), dtype=bool)
    mask = mask.reshape(batch + batch)

    assert (J[mask] > 0).all()
    assert (J[~mask] == 0).all()



r"""General purpose helpers."""

from __future__ import annotations

__all__ = [
    "Partial",
    "bisection",
    "broadcast",
    "gauss_legendre",
    "odeint",
    "unpack",
]

import math
import numpy as np
import torch
import torch.nn as nn

from functools import lru_cache
from torch import Size, Tensor
from torch.autograd.function import once_differentiable
from typing import Any, Callable, Dict, Iterable, List, Sequence, Tuple, Union


class Partial(nn.Module):
    r"""A version of :class:`functools.partial` that is a :class:`torch.nn.Module`.

    Arguments:
        f: An arbitrary callable. If `f` is a module, it is registered as a submodule.
        args: The positional arguments passed to `f`.
        buffer: Whether tensor arguments are registered as buffers or parameters.
        kwargs: The keyword arguments passed to `f`.

    Examples:
        >>> increment = Partial(torch.add, torch.tensor(1.0), buffer=True)
        >>> increment(torch.arange(3))
        tensor([1., 2., 3.])

        >>> weight = torch.randn((5, 3))
        >>> linear = Partial(torch.nn.functional.linear, weight=weight)
        >>> x = torch.rand(2, 3)
        >>> linear(x)
        tensor([[-0.1364, -0.4034,  0.1887, -0.2045, -0.0151],
                [-2.0380, -1.5081, -0.4816,  0.0323, -0.7941]], grad_fn=<MmBackward0>)

        >>> f = torch.distributions.Normal
        >>> loc, scale = torch.zeros(3), torch.ones(3)
        >>> dist = Partial(f, loc, scale, buffer=True)
        >>> dist()
        Normal(loc: torch.Size([3]), scale: torch.Size([3]))
        >>> dist().sample()
        tensor([ 0.1227,  0.1494, -0.6709])
    """

    def __init__(
        self,
        f: Callable,
        /,
        *args: Any,
        buffer: bool = False,
        **kwargs: Any,
    ):
        super().__init__()

        self.f = f

        for i, arg in enumerate(args):
            if torch.is_tensor(arg):
                if buffer:
                    self.register_buffer(f"_{i}", arg)
                else:
                    self.register_parameter(f"_{i}", nn.Parameter(arg))
            else:
                setattr(self, f"_{i}", arg)

        self._nargs = len(args)

        for key, arg in kwargs.items():
            if torch.is_tensor(arg):
                if buffer:
                    self.register_buffer(key, arg)
                else:
                    self.register_parameter(key, nn.Parameter(arg))
            else:
                setattr(self, key, arg)

        self._keys = list(kwargs.keys())

    @property
    def args(self) -> Sequence[Any]:
        return [getattr(self, f"_{i}") for i in range(self._nargs)]

    @property
    def kwargs(self) -> Dict[str, Any]:
        return {key: getattr(self, key) for key in self._keys}

    def extra_repr(self) -> str:
        if isinstance(self.f, nn.Module):
            return ""
        else:
            return f"(f): {self.f}"

    def forward(self, *args, **kwargs) -> Any:
        r"""
        Returns:
            :py:`self.f(*self.args, *args, **self.kwargs, **kwargs)`
        """

        return self.f(
            *self.args,
            *args,
            **self.kwargs,
            **kwargs,
        )


def bisection(
    f: Callable[[Tensor], Tensor],
    y: Tensor,
    a: Union[float, Tensor],
    b: Union[float, Tensor],
    n: int = 16,
    phi: Iterable[Tensor] = (),
) -> Tensor:
    r"""Applies the bisection method to find :math:`x` between the bounds :math:`a`
    and :math:`b` such that :math:`f_\phi(x)` is close to :math:`y`.

    Gradients are propagated through :math:`y` and :math:`\phi` via implicit
    differentiation.

    Wikipedia:
        https://wikipedia.org/wiki/Bisection_method

    Arguments:
        f: A univariate function :math:`f_\phi`.
        y: The target :math:`y`.
        a: The bound :math:`a` such that :math:`f_\phi(a) \leq y`.
        b: The bound :math:`b` such that :math:`y \leq f_\phi(b)`.
        n: The number of iterations.
        phi: The parameters :math:`\phi` of :math:`f_\phi`.

    Returns:
        The solution :math:`x`.

    Example:
        >>> f = torch.cos
        >>> y = torch.tensor(0.0)
        >>> bisection(f, y, 2.0, 1.0, n=16)
        tensor(1.5708)
    """

    a = torch.as_tensor(a).to(y)
    b = torch.as_tensor(b).to(y)

    return Bisection.apply(f, y, a, b, n, *phi)


class Bisection(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        f: Callable[[Tensor], Tensor],
        y: Tensor,
        a: Tensor,
        b: Tensor,
        n: int,
        *phi: Tensor,
    ) -> Tensor:
        for _ in range(n):
            c = (a + b) / 2

            mask = f(c) < y

            a = torch.where(mask, c, a)
            b = torch.where(mask, b, c)

        x = (a + b) / 2

        ctx.f = f
        ctx.save_for_backward(x, *phi)

        return x

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_x: Tensor) -> Tuple[Tensor, ...]:
        f = ctx.f
        x, *phi = ctx.saved_tensors

        with torch.enable_grad():
            x = x.detach().requires_grad_()
            y = f(x)

        (jacobian,) = torch.autograd.grad(y, x, torch.ones_like(y), retain_graph=True)
        grad_y = grad_x / jacobian

        if phi:
            grad_phi = torch.autograd.grad(y, phi, -grad_y, retain_graph=True)
        else:
            grad_phi = ()

        return (None, grad_y, None, None, None, *grad_phi)


def broadcast(*tensors: Tensor, ignore: Union[int, Sequence[int]] = 0) -> List[Tensor]:
    r"""Broadcasts tensors together.

    The term broadcasting describes how PyTorch treats tensors with different shapes
    during arithmetic operations. In short, if possible, dimensions that have
    different sizes are expanded (without making copies) to be compatible.

    Arguments:
        tensors: The tensors to broadcast.
        ignore: The number(s) of dimensions not to broadcast.

    Returns:
        The broadcasted tensors.

    Example:
        >>> x = torch.rand(3, 1, 2)
        >>> y = torch.rand(4, 5)
        >>> x, y = broadcast(x, y, ignore=1)
        >>> x.shape
        torch.Size([3, 4, 2])
        >>> y.shape
        torch.Size([3, 4, 5])
    """

    if isinstance(ignore, int):
        ignore = [ignore] * len(tensors)

    dims = [t.dim() - i for t, i in zip(tensors, ignore)]
    common = torch.broadcast_shapes(*(t.shape[:i] for t, i in zip(tensors, dims)))

    return [torch.broadcast_to(t, common + t.shape[i:]) for t, i in zip(tensors, dims)]


def gauss_legendre(
    f: Callable[[Tensor], Tensor],
    a: Tensor,
    b: Tensor,
    n: int = 3,
    phi: Iterable[Tensor] = (),
) -> Tensor:
    r"""Estimates the definite integral of a function :math:`f_\phi(x)` from :math:`a`
    to :math:`b` using a :math:`n`-point Gauss-Legendre quadrature.

    .. math:: \int_a^b f_\phi(x) ~ dx \approx (b - a) \sum_{i = 1}^n w_i f_\phi(x_i)

    Wikipedia:
        https://wikipedia.org/wiki/Gauss-Legendre_quadrature

    Arguments:
        f: A univariate function :math:`f_\phi`.
        a: The lower limit :math:`a`.
        b: The upper limit :math:`b`.
        n: The number of points :math:`n` at which the function is evaluated.
        phi: The parameters :math:`\phi` of :math:`f_\phi`.

    Returns:
        The definite integral estimation.

    Example:
        >>> f = lambda x: torch.exp(-x**2)
        >>> a, b = torch.tensor([-0.69, 4.2])
        >>> gauss_legendre(f, a, b, n=16)
        tensor(1.4807)
    """

    return GaussLegendre.apply(f, a, b, n, *phi)


class GaussLegendre(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        f: Callable[[Tensor], Tensor],
        a: Tensor,
        b: Tensor,
        n: int,
        *phi: Tensor,
    ) -> Tensor:
        ctx.f, ctx.n = f, n
        ctx.save_for_backward(a, b, *phi)

        return GaussLegendre.quadrature(f, a, b, n)

    @staticmethod
    def backward(ctx, grad_area: Tensor) -> Tuple[Tensor, ...]:
        f, n = ctx.f, ctx.n
        a, b, *phi = ctx.saved_tensors

        if ctx.needs_input_grad[1]:
            grad_a = -f(a) * grad_area
        else:
            grad_a = None

        if ctx.needs_input_grad[2]:
            grad_b = f(b) * grad_area
        else:
            grad_b = None

        if phi:
            with torch.enable_grad():
                area = GaussLegendre.quadrature(f, a, b, n)

            grad_phi = torch.autograd.grad(area, phi, grad_area, create_graph=True)
        else:
            grad_phi = ()

        return (None, grad_a, grad_b, None, *grad_phi)

    @staticmethod
    @lru_cache(maxsize=None)
    def nodes(n: int, **kwargs) -> Tuple[Tensor, Tensor]:
        r"""Returns the nodes and weights for a :math:`n`-point Gauss-Legendre
        quadrature over the interval :math:`[0, 1]`.

        See :func:`numpy.polynomial.legendre.leggauss`.
        """

        nodes, weights = np.polynomial.legendre.leggauss(n)

        nodes = (nodes + 1) / 2
        weights = weights / 2

        kwargs.setdefault("dtype", torch.get_default_dtype())

        return (
            torch.as_tensor(nodes, **kwargs),
            torch.as_tensor(weights, **kwargs),
        )

    @staticmethod
    def quadrature(
        f: Callable[[Tensor], Tensor],
        a: Tensor,
        b: Tensor,
        n: int,
    ) -> Tensor:
        nodes, weights = GaussLegendre.nodes(n, dtype=a.dtype, device=a.device)
        nodes = torch.lerp(
            a[..., None],
            b[..., None],
            nodes,
        ).movedim(-1, 0)

        return (b - a) * torch.tensordot(weights, f(nodes), dims=1)


def odeint(
    f: Callable[[Tensor, Tensor], Tensor],
    x: Union[Tensor, Sequence[Tensor]],
    t0: Union[float, Tensor],
    t1: Union[float, Tensor],
    phi: Iterable[Tensor] = (),
    atol: float = 1e-6,
    rtol: float = 1e-5,
) -> Union[Tensor, Sequence[Tensor]]:
    r"""Integrates a system of first-order ordinary differential equations (ODEs)

    .. math:: \frac{dx}{dt} = f_\phi(t, x) ,

    from :math:`t_0` to :math:`t_1` using the adaptive Dormand-Prince method. The
    output is the final state

    .. math:: x(t_1) = x_0 + \int_{t_0}^{t_1} f_\phi(t, x(t)) ~ dt .

    Gradients are propagated through :math:`x_0`, :math:`t_0`, :math:`t_1` and
    :math:`\phi` via the adaptive checkpoint adjoint (ACA) method.

    References:
        | Neural Ordinary Differential Equations (Chen el al., 2018)
        | https://arxiv.org/abs/1806.07366

        | Adaptive Checkpoint Adjoint Method for Gradient Estimation in Neural ODE (Zhuang et al., 2020)
        | https://arxiv.org/abs/2006.02493

    Arguments:
        f: A system of first-order ODEs :math:`f_\phi`.
        x: The initial state :math:`x_0`.
        t0: The initial integration time :math:`t_0`.
        t1: The final integration time :math:`t_1`.
        phi: The parameters :math:`\phi` of :math:`f_\phi`.
        atol: The absolute tolerance.
        rtol: The relative tolerance.

    Returns:
        The final state :math:`x(t_1)`.

    Example:
        >>> A = torch.randn(3, 3)
        >>> f = lambda t, x: x @ A
        >>> x0 = torch.randn(3)
        >>> x1 = odeint(f, x0, 0.0, 1.0)
        >>> x1
        tensor([-1.4596,  0.5008,  1.5828])
    """

    settings = (atol, rtol, torch.is_grad_enabled())

    if torch.is_tensor(x):
        x0 = x
        g = f
    else:
        shapes = [y.shape for y in x]

        def pack(x: Iterable[Tensor]) -> Tensor:
            return torch.cat([y.flatten() for y in x])

        x0 = pack(x)
        g = lambda t, x: pack(f(t, *unpack(x, shapes)))

    t0 = torch.as_tensor(t0, dtype=x0.dtype, device=x0.device)
    t1 = torch.as_tensor(t1, dtype=x0.dtype, device=x0.device)

    assert not t0.shape and not t1.shape, "'t0' and 't1' must be scalars"

    x1 = AdaptiveCheckpointAdjoint.apply(settings, g, x0, t0, t1, *phi)

    if torch.is_tensor(x):
        return x1
    else:
        return unpack(x1, shapes)


# fmt: off
def dopri45(
    f: Callable[[Tensor, Tensor], Tensor],
    x: Tensor,
    t: Tensor,
    dt: Tensor,
    error: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    r"""Applies one step of the Dormand-Prince method.

    Wikipedia:
        https://wikipedia.org/wiki/Dormand-Prince_method
    """

    k1 = dt * f(t, x)
    k2 = dt * f(t + 1 / 5 * dt, x + 1 / 5 * k1)
    k3 = dt * f(t + 3 / 10 * dt, x + 3 / 40 * k1 + 9 / 40 * k2)
    k4 = dt * f(t + 4 / 5 * dt, x + 44 / 45 * k1 - 56 / 15 * k2 + 32 / 9 * k3)
    k5 = dt * f(
        t + 8 / 9 * dt,
        x + 19372 / 6561 * k1 - 25360 / 2187 * k2 + 64448 / 6561 * k3 - 212 / 729 * k4,
    )
    k6 = dt * f(
        t + dt,
        x
        + 9017 / 3168 * k1
        - 355 / 33 * k2
        + 46732 / 5247 * k3
        + 49 / 176 * k4
        - 5103 / 18656 * k5,
    )
    x_next = (
        x
        + 35 / 384 * k1
        + 500 / 1113 * k3
        + 125 / 192 * k4
        - 2187 / 6784 * k5
        + 11 / 84 * k6
    )

    if not error:
        return x_next

    k7 = dt * f(t + dt, x_next)
    x_star = (
        x
        + 5179 / 57600 * k1
        + 7571 / 16695 * k3
        + 393 / 640 * k4
        - 92097 / 339200 * k5
        + 187 / 2100 * k6
        + 1 / 40 * k7
    )

    return x_next, abs(x_next - x_star)
# fmt: on


class NestedTensor(tuple):
    r"""Creates an efficient data structure to hold and perform basic operations
    on sequences of tensors.
    """

    def __add__(self, other: NestedTensor) -> NestedTensor:
        return NestedTensor(x + y for x, y in zip(self, other))

    def __sub__(self, other: NestedTensor) -> NestedTensor:
        return NestedTensor(x - y for x, y in zip(self, other))

    def __rmul__(self, other: Tensor) -> NestedTensor:
        return NestedTensor(other * x for x in self)


class AdaptiveCheckpointAdjoint(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        settings: Tuple[float, float, bool],
        f: Callable[[Tensor, Tensor], Tensor],
        x: Tensor,
        t0: Tensor,
        t1: Tensor,
        *phi: Tensor,
    ) -> Tensor:
        atol, rtol, grad_enabled = settings

        ctx.f = f
        ctx.save_for_backward(x, t0, t1, *phi)
        ctx.steps = []

        t, dt = t0, t1 - t0
        sign = torch.sign(dt)

        while sign * (t1 - t) > 0:
            dt = sign * torch.min(abs(dt), abs(t1 - t))

            while True:
                y, error = dopri45(f, x, t, dt, error=True)
                tolerance = atol + rtol * torch.max(abs(x), abs(y))
                error = torch.max(error / tolerance).clip(min=1e-9).item()

                if error < 1.0:
                    x, t = y, t + dt

                    if grad_enabled:
                        ctx.steps.append((x, t, dt))

                dt = dt * min(10.0, max(0.1, 0.9 / error ** (1 / 5)))

                if error < 1.0:
                    break

        return x

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_x: Tensor) -> Tuple[Tensor, ...]:
        f = ctx.f
        x0, t0, t1, *phi = ctx.saved_tensors
        x1, _, _ = ctx.steps[-1]

        # Final time
        if ctx.needs_input_grad[4]:
            grad_t1 = torch.sum(f(t1, x1) * grad_x)
        else:
            grad_t1 = None

        # Adjoint
        grad_phi = map(torch.zeros_like, phi)

        def g(t: Tensor, x_aug: NestedTensor) -> NestedTensor:
            x, grad_x, *_ = x_aug

            with torch.enable_grad():
                x = x.detach().requires_grad_()
                dx = f(t, x)

            grad_x, *grad_phi = torch.autograd.grad(dx, (x, *phi), -grad_x)

            return NestedTensor((dx, grad_x, *grad_phi))

        for x, t, dt in reversed(ctx.steps):
            x_aug = NestedTensor((x, grad_x, *grad_phi))
            _, grad_x, *grad_phi = dopri45(g, x_aug, t, -dt)

        # Initial time
        if ctx.needs_input_grad[3]:
            grad_t0 = torch.sum(f(t0, x0) * grad_x)
        else:
            grad_t0 = None

        return (None, None, grad_x, grad_t0, grad_t1, *grad_phi)


def unpack(x: Tensor, shapes: Sequence[Size]) -> Sequence[Tensor]:
    r"""Unpacks a packed tensor.

    Arguments:
        x: A packed tensor, with shape :math:`(*, D)`.
        shapes: A sequence of shapes :math:`S_i`, corresponding to the total number of
            elements :math:`D`.

    Returns:
        The unpacked tensors, with shapes :math:`(*, S_i)`.

    Example:
        >>> x = torch.randn(26)
        >>> y, z = unpack(x, ((1, 2, 3), (4, 5)))
        >>> y.shape
        torch.Size([1, 2, 3])
        >>> z.shape
        torch.Size([4, 5])
    """

    sizes = [math.prod(s) for s in shapes]

    x = x.split(sizes, -1)
    x = (y.unflatten(-1, (*s, 1)) for y, s in zip(x, shapes))
    x = (y.squeeze(-1) for y in x)

    return tuple(x)



r"""Neural flows and transformations."""

__all__ = [
    "MNN",
    "NAF",
    "UMNN",
    "UNAF",
]

import torch
import torch.nn as nn

from functools import partial
from torch import Tensor
from torch.distributions import Transform
from typing import Any, Dict

from .autoregressive import MaskedAutoregressiveTransform
from ..distributions import DiagNormal
from ..lazy import Flow, UnconditionalDistribution, UnconditionalTransform
from ..nn import MLP, MonotonicMLP
from ..transforms import (
    AdditiveTransform,
    ComposedTransform,
    MonotonicTransform,
    SoftclipTransform,
    UnconstrainedMonotonicTransform,
)
from ..utils import broadcast


class MNN(nn.Module):
    r"""Creates a monotonic neural network (MNN).

    The monotonic neural network is parametrized by its internal positive weights, which
    are independent of the features and context. To modulate its behavior, it receives
    as input a signal vector that can depend on the features and context.

    See also:
        :class:`zuko.transforms.MonotonicTransform`

    References:
        | Neural Autoregressive Flows (Huang et al., 2018)
        | https://arxiv.org/abs/1804.00779

    Arguments:
        signal: The number of signal features of the monotonic network.
        kwargs: Keyword arguments passed to :class:`zuko.nn.MonotonicMLP`.
    """

    def __init__(self, signal: int = 16, **kwargs):
        super().__init__()

        self.network = MonotonicMLP(1 + signal, 1, **kwargs)

    def f(self, signal: Tensor, x: Tensor) -> Tensor:
        y = self.network(torch.cat(broadcast(x[..., None], signal, ignore=1), dim=-1))
        y = y.squeeze(dim=-1)

        return y

    def forward(self, signal: Tensor) -> Transform:
        return MonotonicTransform(
            f=partial(self.f, signal),
            phi=(signal, *self.parameters()),
        )


class UMNN(nn.Module):
    r"""Creates an unconstrained monotonic neural network (UMNN).

    The integrand neural network is parametrized by its internal weights, which are
    independent of the features and context. To modulate its behavior, it receives as
    input a signal vector that can depend on the features and context.

    See also:
        :class:`zuko.transforms.UnconstrainedMonotonicTransform`

    References:
        | Unconstrained Monotonic Neural Networks (Wehenkel et al., 2019)
        | https://arxiv.org/abs/1908.05164

    Arguments:
        signal: The number of signal features of the integrand network.
        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.
    """

    def __init__(self, signal: int = 16, **kwargs):
        super().__init__()

        kwargs.setdefault("activation", nn.ELU)

        self.integrand = MLP(1 + signal, 1, **kwargs)

    def g(self, signal: Tensor, x: Tensor) -> Tensor:
        dx = self.integrand(torch.cat(broadcast(x[..., None], signal, ignore=1), dim=-1))
        dx = dx.squeeze(dim=-1)

        return torch.exp(dx / (1 + abs(dx / 7)))  # in [1e-3, 1e3]

    def forward(self, signal: Tensor, constant: Tensor) -> Transform:
        return ComposedTransform(
            UnconstrainedMonotonicTransform(
                g=partial(self.g, signal),
                phi=(signal, *self.parameters()),
            ),
            AdditiveTransform(shift=constant),
        )


class NAF(Flow):
    r"""Creates a neural autoregressive flow (NAF).

    Warning:
        Invertibility is only guaranteed for features within the interval :math:`[-10,
        10]`. It is recommended to standardize features (zero mean, unit variance)
        before training.

    References:
        | Neural Autoregressive Flows (Huang et al., 2018)
        | https://arxiv.org/abs/1804.00779

    Arguments:
        features: The number of features.
        context: The number of context features.
        transforms: The number of autoregressive transformations.
        randperm: Whether features are randomly permuted between transformations or not.
            If :py:`False`, features are in ascending (descending) order for even
            (odd) transformations.
        signal: The number of signal features of the monotonic network.
        network: Keyword arguments passed to :class:`MNN`.
        kwargs: Keyword arguments passed to :class:`zuko.flows.autoregressive.MaskedAutoregressiveTransform`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        transforms: int = 3,
        randperm: bool = False,
        signal: int = 16,
        network: Dict[str, Any] = {},  # noqa: B006
        **kwargs,
    ):
        orders = [
            torch.arange(features),
            torch.flipud(torch.arange(features)),
        ]

        transforms = [
            MaskedAutoregressiveTransform(
                features=features,
                context=context,
                order=torch.randperm(features) if randperm else orders[i % 2],
                univariate=MNN(signal=signal, stack=features, **network),
                shapes=[(signal,)],
                **kwargs,
            )
            for i in range(transforms)
        ]

        for i in reversed(range(1, len(transforms))):
            transforms.insert(i, UnconditionalTransform(SoftclipTransform, bound=11.0))

        base = UnconditionalDistribution(
            DiagNormal,
            torch.zeros(features),
            torch.ones(features),
            buffer=True,
        )

        super().__init__(transforms, base)


class UNAF(Flow):
    r"""Creates an unconstrained neural autoregressive flow (UNAF).

    Warning:
        Invertibility is only guaranteed for features within the interval :math:`[-10,
        10]`. It is recommended to standardize features (zero mean, unit variance)
        before training.

    References:
        | Unconstrained Monotonic Neural Networks (Wehenkel et al., 2019)
        | https://arxiv.org/abs/1908.05164

    Arguments:
        features: The number of features.
        context: The number of context features.
        transforms: The number of autoregressive transformations.
        randperm: Whether features are randomly permuted between transformations or not.
            If :py:`False`, features are in ascending (descending) order for even
            (odd) transformations.
        signal: The number of signal features of the monotonic network.
        network: Keyword arguments passed to :class:`UMNN`.
        kwargs: Keyword arguments passed to :class:`zuko.flows.autoregressive.MaskedAutoregressiveTransform`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        transforms: int = 3,
        randperm: bool = False,
        signal: int = 16,
        network: Dict[str, Any] = {},  # noqa: B006
        **kwargs,
    ):
        orders = [
            torch.arange(features),
            torch.flipud(torch.arange(features)),
        ]

        transforms = [
            MaskedAutoregressiveTransform(
                features=features,
                context=context,
                order=torch.randperm(features) if randperm else orders[i % 2],
                univariate=UMNN(signal=signal, stack=features, **network),
                shapes=[(signal,), ()],
                **kwargs,
            )
            for i in range(transforms)
        ]

        for i in reversed(range(1, len(transforms))):
            transforms.insert(i, UnconditionalTransform(SoftclipTransform, bound=11.0))

        base = UnconditionalDistribution(
            DiagNormal,
            torch.zeros(features),
            torch.ones(features),
            buffer=True,
        )

        super().__init__(transforms, base)



r"""Continuous flows and transformations."""

__all__ = [
    "CNF",
    "FFJTransform",
]

import torch
import torch.nn as nn

from functools import partial
from math import pi
from torch import Tensor
from torch.distributions import Transform

from ..distributions import DiagNormal
from ..lazy import Flow, LazyTransform, UnconditionalDistribution
from ..nn import MLP
from ..transforms import FreeFormJacobianTransform
from ..utils import broadcast


class FFJTransform(LazyTransform):
    r"""Creates a lazy free-form Jacobian (FFJ) transformation.

    See also:
        :class:`zuko.transforms.FreeFormJacobianTransform`

    References:
        | FFJORD: Free-form Continuous Dynamics for Scalable Reversible Generative Models (Grathwohl et al., 2018)
        | https://arxiv.org/abs/1810.01367

    Arguments:
        features: The number of features.
        context: The number of context features.
        freqs: The number of time embedding frequencies.
        atol: The absolute integration tolerance.
        rtol: The relative integration tolerance.
        exact: Whether the exact log-determinant of the Jacobian or an unbiased
            stochastic estimate thereof is calculated.
        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.

    Example:
        >>> t = FFJTransform(3, 4)
        >>> t
        FFJTransform(
          (ode): MLP(
            (0): Linear(in_features=13, out_features=64, bias=True)
            (1): ELU(alpha=1.0)
            (2): Linear(in_features=64, out_features=64, bias=True)
            (3): ELU(alpha=1.0)
            (4): Linear(in_features=64, out_features=3, bias=True)
          )
        )
        >>> x = torch.randn(3)
        >>> x
        tensor([ 0.6365, -0.3181,  1.1519])
        >>> c = torch.randn(4)
        >>> y = t(c)(x)
        >>> t(c).inv(y)
        tensor([ 0.6364, -0.3181,  1.1519],
               grad_fn=<AdaptiveCheckpointAdjointBackward>)
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        freqs: int = 3,
        atol: float = 1e-6,
        rtol: float = 1e-5,
        exact: bool = True,
        **kwargs,
    ):
        super().__init__()

        kwargs.setdefault("activation", nn.ELU)

        self.ode = MLP(features + context + 2 * freqs, features, **kwargs)

        self.register_buffer("times", torch.tensor((0.0, 1.0)))
        self.register_buffer("freqs", torch.arange(1, freqs + 1) * pi)

        self.atol = atol
        self.rtol = rtol
        self.exact = exact

    def f(self, t: Tensor, x: Tensor, c: Tensor = None) -> Tensor:
        t = self.freqs * t[..., None]
        t = torch.cat((t.cos(), t.sin()), dim=-1)

        if c is None:
            x = torch.cat(broadcast(t, x, ignore=1), dim=-1)
        else:
            x = torch.cat(broadcast(t, x, c, ignore=1), dim=-1)

        return self.ode(x)

    def forward(self, c: Tensor = None) -> Transform:
        return FreeFormJacobianTransform(
            f=partial(self.f, c=c),
            t0=self.times[0],
            t1=self.times[1],
            phi=self.parameters() if c is None else (c, *self.parameters()),
            atol=self.atol,
            rtol=self.rtol,
            exact=self.exact,
        )


class CNF(Flow):
    r"""Creates a continuous normalizing flow (CNF) with a free-form Jacobian
    transformation.

    References:
        | Neural Ordinary Differential Equations (Chen el al., 2018)
        | https://arxiv.org/abs/1806.07366

        | FFJORD: Free-form Continuous Dynamics for Scalable Reversible Generative Models (Grathwohl et al., 2018)
        | https://arxiv.org/abs/1810.01367

    Arguments:
        features: The number of features.
        context: The number of context features.
        kwargs: Keyword arguments passed to :class:`FFJTransform`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        **kwargs,
    ):
        transform = FFJTransform(
            features=features,
            context=context,
            **kwargs,
        )

        base = UnconditionalDistribution(
            DiagNormal,
            torch.zeros(features),
            torch.ones(features),
            buffer=True,
        )

        super().__init__(transform, base)
